@@ -6,13 +6,20 @@
     security via Flask-Login, Flask-Principal, Flask-WTF, and passlib.
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner.
+    :copyright: (c) 2019-2022 by J. Christopher Wagner.
     :license: MIT, see LICENSE for more details.
 """
 
 # flake8: noqa: F401
 from .changeable import admin_change_password
-from .core import Security, RoleMixin, UserMixin, AnonymousUser, current_user
+from .core import (
+    Security,
+    RoleMixin,
+    UserMixin,
+    WebAuthnMixin,
+    AnonymousUser,
+    current_user,
+)
 from .datastore import (
     UserDatastore,
     SQLAlchemyUserDatastore,
@@ -42,6 +49,8 @@ from .forms import (
     ResetPasswordForm,
     PasswordlessLoginForm,
     ConfirmRegisterForm,
+    MfRecoveryForm,
+    MfRecoveryCodesForm,
     SendConfirmationForm,
     TwoFactorRescueForm,
     TwoFactorSetupForm,
@@ -67,9 +76,12 @@ from .signals import (
     user_registered,
     us_security_token_sent,
     us_profile_changed,
+    wan_deleted,
+    wan_registered,
 )
 from .totp import Totp
 from .twofactor import tf_send_security_token
+from .tf_plugin import TwoFactorSelectForm
 from .unified_signin import (
     UnifiedSigninForm,
     UnifiedSigninSetupForm,
@@ -79,7 +91,6 @@ from .unified_signin import (
 )
 from .username_util import UsernameUtil
 from .utils import (
-    FsJsonEncoder,
     SmsSenderBaseClass,
     SmsSenderFactory,
     check_and_get_token_status,
@@ -91,6 +102,7 @@ from .utils import (
     check_and_update_authn_fresh,
     login_user,
     logout_user,
+    lookup_identity,
     password_breached_validator,
     password_complexity_validator,
     password_length_validator,
@@ -104,5 +116,14 @@ from .utils import (
     verify_password,
     verify_and_update_password,
 )
+from .webauthn import (
+    WebAuthnRegisterForm,
+    WebAuthnRegisterResponseForm,
+    WebAuthnSigninForm,
+    WebAuthnSigninResponseForm,
+    WebAuthnDeleteForm,
+    WebAuthnVerifyForm,
+)
+from .webauthn_util import WebauthnUtil
 
-__version__ = "4.1.1"
+__version__ = "5.0.0"
