@@ -5,7 +5,7 @@
     Command Line Interface for managing accounts and roles.
 
     :copyright: (c) 2016 by CERN.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner
+    :copyright: (c) 2019-2022 by J. Christopher Wagner
     :license: MIT, see LICENSE for more details.
 """
 
@@ -20,7 +20,7 @@ from .quart_compat import get_quart_status
 
 from .changeable import admin_change_password
 from .utils import (
-    find_user,
+    lookup_identity,
     get_identity_attributes,
     get_identity_attribute,
     hash_password,
@@ -44,7 +44,6 @@ if get_quart_status():  # pragma: no cover
 
         return functools.update_wrapper(decorator, f)
 
-
 else:
     import flask.cli
 
@@ -176,7 +175,7 @@ def roles_add(user, role):
 
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
 
@@ -202,7 +201,7 @@ def roles_remove(user, role):
 
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
 
@@ -232,7 +231,8 @@ def roles_add_permissions(role, permissions):
     role = _datastore._prepare_role_modify_args(role)
     if role is None:
         raise click.UsageError("Cannot find role.")
-    if _datastore.add_permissions_to_role(role, permissions):
+    permlist = [s.strip() for s in permissions.split(",")]
+    if _datastore.add_permissions_to_role(role, permlist):
         click.secho(
             f'Permission(s) "{permissions}" added to role "{role.name}" successfully.',
             fg="green",
@@ -255,7 +255,8 @@ def roles_remove_permissions(role, permissions):
     role = _datastore._prepare_role_modify_args(role)
     if role is None:
         raise click.UsageError("Cannot find role.")
-    if _datastore.remove_permissions_from_role(role, permissions):
+    permlist = [s.strip() for s in permissions.split(",")]
+    if _datastore.remove_permissions_from_role(role, permlist):
         click.secho(
             f'Permission(s) "{permissions}" removed from role'
             f' "{role.name}" successfully.',
@@ -274,7 +275,7 @@ def users_activate(user):
 
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
     if _datastore.activate_user(user_obj):
@@ -292,7 +293,7 @@ def users_deactivate(user):
 
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
     if _datastore.deactivate_user(user_obj):
@@ -313,7 +314,7 @@ def users_reset_access(user):
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
 
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
     _datastore.reset_user_access(user_obj)
@@ -332,12 +333,12 @@ def users_change_password(user, password):
     Administratively change a user's password.
     All the user's sessions will be immediately invalidated.
     You will have to inform the user via an out of band mechanism
-    what the their new password is.
+    what their new password is.
 
     USER is identity as defined by SECURITY_USER_IDENTITY_ATTRIBUTES.
 
     """
-    user_obj = find_user(user)
+    user_obj = lookup_identity(user)
     if user_obj is None:
         raise click.UsageError("User not found.")
 
