@@ -5,7 +5,7 @@
     This module contains an user datastore classes.
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 import datetime
@@ -13,7 +13,7 @@ import json
 import typing as t
 import uuid
 
-from .utils import config_value
+from .utils import config_value as cv
 
 if t.TYPE_CHECKING:  # pragma: no cover
     import flask_sqlalchemy
@@ -126,17 +126,37 @@ class UserDatastore:
 
     :param user_model: A user model class definition
     :param role_model: A role model class definition
+    :param webauthn_model: A model used to store webauthn registrations
 
     .. important::
         For mutating operations, the user/role will be added to the
         datastore (by calling self.put(<object>). If the datastore is session based
         (such as for SQLAlchemyDatastore) it is up to caller to actually
         commit the transaction by calling datastore.commit().
+
+
+    .. note::
+        You must implement get_user_mapping in your WebAuthn model
+        if your User model doesn't have a primary key Column called 'id'
     """
 
-    def __init__(self, user_model: t.Type["User"], role_model: t.Type["Role"]):
+    def __init__(
+        self,
+        user_model: t.Type["User"],
+        role_model: t.Type["Role"],
+        webauthn_model: t.Optional[t.Type["WebAuthn"]] = None,
+    ):
         self.user_model = user_model
         self.role_model = role_model
+        self.webauthn_model = webauthn_model
+
+    if t.TYPE_CHECKING:  # pragma: no cover
+        # These are available from a DataStore implementation
+        def delete(self, model):
+            pass
+
+        def put(self, model):
+            pass
 
     def _prepare_role_modify_args(
         self, role: t.Union[str, "Role"]
@@ -156,13 +176,16 @@ class UserDatastore:
         kwargs.setdefault("fs_uniquifier", uuid.uuid4().hex)
         if hasattr(self.user_model, "fs_token_uniquifier"):
             kwargs.setdefault("fs_token_uniquifier", uuid.uuid4().hex)
+        if hasattr(self.user_model, "fs_webauthn_user_handle"):
+            kwargs.setdefault("fs_webauthn_user_handle", uuid.uuid4().hex)
 
         return kwargs
 
-    def find_user(
-        self, case_insensitive: bool = False, **kwargs: t.Any
-    ) -> t.Union["User", None]:
-        """Returns a user matching the provided parameters."""
+    def find_user(self, **kwargs: t.Any) -> t.Union["User", None]:
+        """Returns a user matching the provided parameters.
+        Besides keyword arguments used to filter the results,
+        'case_insensitive' can be passed (defaults to False)
+        """
         raise NotImplementedError
 
     def find_role(self, role: str) -> t.Union["Role", None]:
@@ -182,7 +205,7 @@ class UserDatastore:
             raise ValueError(f"Role: {role} doesn't exist")
         if role_obj not in user.roles:
             user.roles.append(role_obj)
-            self.put(user)  # type: ignore
+            self.put(user)
             return True
         return False
 
@@ -200,17 +223,17 @@ class UserDatastore:
         if role_obj in user.roles:
             rv = True
             user.roles.remove(role_obj)
-            self.put(user)  # type: ignore
+            self.put(user)
         return rv
 
     def add_permissions_to_role(
-        self, role: t.Union["Role", str], permissions: t.Union[set, list, str]
+        self, role: t.Union["Role", str], permissions: t.Union[set, list, tuple, str]
     ) -> bool:
         """Add one or more permissions to role.
 
         :param role: The role to modify. Can be a Role object or
             string role name
-        :param permissions: a set, list, or single string.
+        :param permissions: a set, list, tuple or comma separated string.
         :return: True if permissions added, False if role doesn't exist.
 
         Caller must commit to DB.
@@ -222,18 +245,24 @@ class UserDatastore:
         role_obj = self._prepare_role_modify_args(role)
         if role_obj:
             rv = True
-            role_obj.add_permissions(permissions)
-            self.put(role_obj)  # type: ignore
+            current_perms = role_obj.get_permissions()
+            if isinstance(permissions, set) or isinstance(permissions, tuple):
+                permissions = list(permissions)
+            elif isinstance(permissions, str):
+                permissions = [p.strip() for p in permissions.split(",")]
+            # always give a list to DB - some (e.g. Mongo) only take list/tuple
+            role_obj.permissions = list(current_perms.union(set(permissions)))
+            self.put(role_obj)
         return rv
 
     def remove_permissions_from_role(
-        self, role: t.Union["Role", str], permissions: t.Union[set, list, str]
+        self, role: t.Union["Role", str], permissions: t.Union[set, list, tuple, str]
     ) -> bool:
         """Remove one or more permissions from a role.
 
         :param role: The role to modify. Can be a Role object or
             string role name
-        :param permissions: a set, list, or single string.
+        :param permissions: a set, list, tuple or a comma separated string.
         :return: True if permissions removed, False if role doesn't exist.
 
         Caller must commit to DB.
@@ -245,14 +274,19 @@ class UserDatastore:
         role_obj = self._prepare_role_modify_args(role)
         if role_obj:
             rv = True
-            role_obj.remove_permissions(permissions)
-            self.put(role_obj)  # type: ignore
+            current_perms = role_obj.get_permissions()
+            if isinstance(permissions, set) or isinstance(permissions, tuple):
+                permissions = list(permissions)
+            elif isinstance(permissions, str):
+                permissions = [p.strip() for p in permissions.split(",")]
+            role_obj.permissions = list(current_perms.difference(set(permissions)))
+            self.put(role_obj)
         return rv
 
     def toggle_active(self, user: "User") -> bool:
         """Toggles a user's active status. Always returns True."""
         user.active = not user.active
-        self.put(user)  # type: ignore
+        self.put(user)
         return True
 
     def deactivate_user(self, user: "User") -> bool:
@@ -266,7 +300,7 @@ class UserDatastore:
         """
         if user.active:
             user.active = False
-            self.put(user)  # type: ignore
+            self.put(user)
             return True
         return False
 
@@ -277,7 +311,7 @@ class UserDatastore:
         """
         if not user.active:
             user.active = True
-            self.put(user)  # type: ignore
+            self.put(user)
             return True
         return False
 
@@ -296,7 +330,7 @@ class UserDatastore:
         if not uniquifier:
             uniquifier = uuid.uuid4().hex
         user.fs_uniquifier = uniquifier
-        self.put(user)  # type: ignore
+        self.put(user)
 
     def set_token_uniquifier(
         self, user: "User", uniquifier: t.Union[str, None] = None
@@ -316,7 +350,7 @@ class UserDatastore:
             uniquifier = uuid.uuid4().hex
         if hasattr(user, "fs_token_uniquifier"):
             user.fs_token_uniquifier = uniquifier
-            self.put(user)  # type: ignore
+            self.put(user)
 
     def create_role(self, **kwargs: t.Any) -> "Role":
         """
@@ -324,27 +358,26 @@ class UserDatastore:
         Supported params (depending on RoleModel):
 
         :kwparam name: Role name
-        :kwparam permissions: a comma delimited list of permissions, a set or a list.
-            These are user-defined strings that correspond to strings used with
+        :kwparam permissions: a list, set, tuple or comma separated string.
+            These are user-defined strings that correspond to args used with
             @permissions_required()
 
             .. versionadded:: 3.3.0
 
         """
 
-        # By default we just use raw DB model create - for permissions we want to
-        # be nicer and allow sending in a list or set or comma separated string.
+        # Usually we just use raw DB model create - for permissions we want to
+        # be nicer and allow sending in a list or set or a single string.
         if "permissions" in kwargs and hasattr(self.role_model, "permissions"):
             perms = kwargs["permissions"]
-            if isinstance(perms, list) or isinstance(perms, set):
-                perms = ",".join(perms)
+            if isinstance(perms, set) or isinstance(perms, tuple):
+                perms = list(perms)
             elif isinstance(perms, str):
-                # squash spaces.
-                perms = ",".join(p.strip() for p in perms.split(","))
+                perms = [p.strip() for p in perms.split(",")]
             kwargs["permissions"] = perms
 
         role = self.role_model(**kwargs)
-        return self.put(role)  # type: ignore
+        return self.put(role)
 
     def find_or_create_role(self, name: str, **kwargs: t.Any) -> "Role":
         """Returns a role matching the given name or creates it with any
@@ -373,7 +406,7 @@ class UserDatastore:
                 except ValueError:
 
         .. note::
-            The roles kwparam is modified as part of the call - it will, if necessary
+            The roles kwparam is modified as part of the call - it will, if necessary,
             be converted from names to role instances.
 
         .. danger::
@@ -396,7 +429,7 @@ class UserDatastore:
         """
         kwargs = self._prepare_create_user_args(**kwargs)
         user = self.user_model(**kwargs)
-        return self.put(user)  # type: ignore
+        return self.put(user)
 
     def delete_user(self, user: "User") -> None:
         """Deletes the specified user.
@@ -415,10 +448,13 @@ class UserDatastore:
             * reset fs_token_uniquifier (if present) - cause auth tokens to be unusable
             * remove all unified signin TOTP secrets so those can't be used
             * remove all two-factor secrets so those can't be used
+            * remove all registered webauthn credentials
+            * remove all one-time recovery codes
+            * will NOT affect password
 
         Note that if using unified sign in and allow 'email' as a way to receive a code;
-        if the email is compromised - login is still possible. To handle this - it
-        is better to deactivate the user.
+        this will also get reset. If the user registered w/o a password then they likely
+        will have no way to authenticate.
 
         Note - this method isn't used directly by Flask-Security - it is provided
         as a helper for an application's administrative needs.
@@ -426,6 +462,9 @@ class UserDatastore:
         Remember to call commit on DB if needed.
 
         .. versionadded:: 3.4.1
+
+        .. versionchanged:: 5.0.0
+            Added webauthn and recovery codes reset.
         """
         self.set_uniquifier(user)
         self.set_token_uniquifier(user)
@@ -433,6 +472,10 @@ class UserDatastore:
             self.us_reset(user)
         if hasattr(user, "tf_primary_method"):
             self.tf_reset(user)
+        if hasattr(user, "webauthn"):
+            self.webauthn_reset(user)
+        if hasattr(user, "mf_recovery_codes"):
+            self.mf_set_recovery_codes(user, None)
 
     def tf_set(
         self,
@@ -466,17 +509,52 @@ class UserDatastore:
             user.tf_phone_number = phone
             changed = True
         if changed:
-            self.put(user)  # type: ignore
+            self.put(user)
 
     def tf_reset(self, user: "User") -> None:
-        """Disable two-factor auth for user
+        """Disable two-factor auth for user.
 
         .. versionadded: 3.4.1
         """
         user.tf_primary_method = None
         user.tf_totp_secret = None
         user.tf_phone_number = None
-        self.put(user)  # type: ignore
+        self.put(user)
+
+    def mf_set_recovery_codes(self, user: "User", rcs: t.Optional[t.List[str]]) -> None:
+        """Set MF recovery codes into user record.
+        Any existing codes will be erased.
+
+        .. danger::
+           Be aware that whatever `passwords` are passed in will
+           be stored directly in the DB. Do NOT pass in a plaintext passwords!
+           Best practice is to pass in ``hash_password(plaintext_password)``.
+
+        .. versionadded: 5.0.0
+        """
+        user.mf_recovery_codes = rcs
+        self.put(user)
+
+    def mf_get_recovery_codes(self, user: "User") -> t.List[str]:
+        codes = getattr(user, "mf_recovery_codes", [])
+        return codes if codes else []
+
+    def mf_delete_recovery_code(self, user: "User", code_to_delete: str) -> bool:
+        """Delete a single recovery code.
+        Recovery codes are single-use - so delete after using!
+
+        Return True if code found and deleted, False otherwise.
+
+        .. versionadded: 5.0.0
+        """
+        if not user.mf_recovery_codes:
+            return False
+        try:
+            user.mf_recovery_codes.remove(code_to_delete)
+            self.put(user)
+            return True
+        except ValueError:
+            return False
 
     def us_get_totp_secrets(self, user: "User") -> t.Dict[str, str]:
         """Return totp secrets.
@@ -518,7 +596,7 @@ class UserDatastore:
 
         To get a totp_secret - use ``app.security._totp_factory.generate_totp_secret()``
 
-        .. versionadded: 3.4.1
+        .. versionadded:: 3.4.1
         """
 
         if totp_secret:
@@ -527,18 +605,112 @@ class UserDatastore:
             self.us_put_totp_secrets(user, totp_secrets)
         if phone and user.us_phone_number != phone:
             user.us_phone_number = phone
-            self.put(user)  # type: ignore
+            self.put(user)
 
-    def us_reset(self, user: "User") -> None:
+    def us_reset(self, user: "User", method: t.Optional[str] = None) -> None:
         """Disable unified sign in for user.
-        Be aware that if "email" is an allowed way to receive codes, they
-        will still work (as totp secrets are generated on the fly).
-        This will disable authenticator app and SMS.
+        This will disable authenticator app and SMS, and email.
+        N.B. if user has no password they may not be able to authenticate at all.
+
+        .. versionadded:: 3.4.1
+
+        .. versionchanged:: 5.0.0
+            Added optional method argument to delete just a single method
 
-        .. versionadded: 3.4.1
         """
-        user.us_totp_secrets = None
-        self.put(user)  # type: ignore
+        if not method:
+            # delete all
+            self.us_put_totp_secrets(user, None)
+            user.us_phone_number = None
+            self.put(user)
+        else:
+            totp_secrets = self.us_get_totp_secrets(user)
+            del totp_secrets[method]
+            self.us_put_totp_secrets(user, totp_secrets)
+            if method == "sms":
+                user.us_phone_number = None
+                self.put(user)
+
+    def us_setup_email(self, user: "User") -> bool:
+        # setup email (if allowed) for user for unified sign in.
+        from .proxies import _security
+
+        if not cv("UNIFIED_SIGNIN") or "email" not in cv("US_ENABLED_METHODS"):
+            return False
+        totp_secrets = self.us_get_totp_secrets(user)
+        totp_secrets["email"] = _security._totp_factory.generate_totp_secret()
+        self.us_put_totp_secrets(user, totp_secrets)
+        return True
+
+    def set_webauthn_user_handle(
+        self, user: "User", user_handle: t.Union[str, None] = None
+    ) -> None:
+        """Set the value for the Relaying Party's (that's us)
+        UserHandle (user.id)
+        If no value is passed in, a UUID is generated.
+        """
+        if not user_handle:
+            user_handle = uuid.uuid4().hex
+        user.fs_webauthn_user_handle = user_handle
+        self.put(user)
+
+    def create_webauthn(
+        self,
+        user: "User",
+        credential_id: bytes,
+        public_key: bytes,
+        name: str,
+        sign_count: int,
+        usage: str,
+        device_type: str,
+        backup_state: bool,
+        transports: t.Optional[t.List[str]] = None,
+        extensions: t.Optional[str] = None,
+        **kwargs: t.Any,
+    ) -> None:
+        """
+        Create a new webauthn registration record.
+        Note that we need to find webauthn records per user as well as
+        find a user from a given webauthn (credential_id) record.
+
+        .. versionadded: 5.0.0
+        """
+        raise NotImplementedError
+
+    def delete_webauthn(self, webauthn: "WebAuthn") -> None:
+        """
+        .. versionadded: 5.0.0
+        """
+        self.delete(webauthn)
+
+    def find_webauthn(self, credential_id: bytes) -> t.Union["WebAuthn", None]:
+        """Returns a credential matching the id.
+
+        .. versionadded: 5.0.0
+        """
+        raise NotImplementedError
+
+    def find_user_from_webauthn(self, webauthn: "WebAuthn") -> t.Union["User", None]:
+        """Returns user associated with this webauthn credential
+
+        .. versionadded: 5.0.0
+        """
+        if not self.webauthn_model:
+            raise NotImplementedError
+        user_filter = webauthn.get_user_mapping()
+        return self.find_user(**user_filter)
+
+    def webauthn_reset(self, user: "User") -> None:
+        """Reset access via webauthn credentials.
+        This will DELETE all registered credentials.
+        There doesn't appear to be any reason to change the user's
+        fs_webauthn_user_handle.
+
+        .. versionadded: 5.0.0
+        """
+        for cred in user.webauthn:
+            self.delete(cred)
+        self.put(user)
 
 
 class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
@@ -550,6 +722,7 @@ class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
     :param db:
     :param user_model: See :ref:`Models <models_topic>`.
     :param role_model: See :ref:`Models <models_topic>`.
+    :param webauthn_model: See :ref:`Models <models_topic>`.
     """
 
     def __init__(
@@ -557,9 +730,10 @@ class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
         db: "flask_sqlalchemy.SQLAlchemy",
         user_model: t.Type["User"],
         role_model: t.Type["Role"],
+        webauthn_model: t.Optional[t.Type["WebAuthn"]] = None,
     ):
         SQLAlchemyDatastore.__init__(self, db)
-        UserDatastore.__init__(self, user_model, role_model)
+        UserDatastore.__init__(self, user_model, role_model, webauthn_model)
 
     def find_user(
         self, case_insensitive: bool = False, **kwargs: t.Any
@@ -567,10 +741,10 @@ class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
         from sqlalchemy import func as alchemyFn
 
         query = self.user_model.query
-        if config_value("JOIN_USER_ROLES") and hasattr(self.user_model, "roles"):
+        if cv("JOIN_USER_ROLES") and hasattr(self.user_model, "roles"):
             from sqlalchemy.orm import joinedload
 
-            query = query.options(joinedload("roles"))
+            query = query.options(joinedload(self.user_model.roles))
 
         if case_insensitive:
             # While it is of course possible to pass in multiple keys to filter on
@@ -589,6 +763,45 @@ class SQLAlchemyUserDatastore(SQLAlchemyDatastore, UserDatastore):
     def find_role(self, role: str) -> t.Union["Role", None]:
         return self.role_model.query.filter_by(name=role).first()  # type: ignore
 
+    def find_webauthn(self, credential_id: bytes) -> t.Union["WebAuthn", None]:
+        return self.webauthn_model.query.filter_by(  # type: ignore
+            credential_id=credential_id
+        ).first()
+
+    def create_webauthn(
+        self,
+        user: "User",
+        credential_id: bytes,
+        public_key: bytes,
+        name: str,
+        sign_count: int,
+        usage: str,
+        device_type: str,
+        backup_state: bool,
+        transports: t.Optional[t.List[str]] = None,
+        extensions: t.Optional[str] = None,
+        **kwargs: t.Any,
+    ) -> None:
+        if not hasattr(self, "webauthn_model") or not self.webauthn_model:
+            raise NotImplementedError
+
+        webauthn = self.webauthn_model(
+            credential_id=credential_id,
+            public_key=public_key,
+            name=name,
+            sign_count=sign_count,
+            usage=usage,
+            device_type=device_type,
+            backup_state=backup_state,
+            transports=transports,
+            extensions=extensions,
+            lastuse_datetime=datetime.datetime.utcnow(),
+            **kwargs,
+        )
+        user.webauthn.append(webauthn)
+        self.put(webauthn)
+        self.put(user)
+
 
 class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastore):
     """A UserDatastore implementation that directly uses
@@ -598,6 +811,7 @@ class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastor
     :param session:
     :param user_model: See :ref:`Models <models_topic>`.
     :param role_model: See :ref:`Models <models_topic>`.
+    :param webauthn_model: See :ref:`Models <models_topic>`.
     """
 
     def __init__(
@@ -605,6 +819,7 @@ class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastor
         session: "sqlalchemy.orm.scoping.scoped_session",
         user_model: t.Type["User"],
         role_model: t.Type["Role"],
+        webauthn_model: t.Optional[t.Type["WebAuthn"]] = None,
     ):
         class PretendFlaskSQLAlchemyDb:
             """This is a pretend db object, so we can just pass in a session."""
@@ -613,7 +828,11 @@ class SQLAlchemySessionUserDatastore(SQLAlchemyUserDatastore, SQLAlchemyDatastor
                 self.session = session
 
         SQLAlchemyUserDatastore.__init__(
-            self, PretendFlaskSQLAlchemyDb(session), user_model, role_model
+            self,
+            PretendFlaskSQLAlchemyDb(session),
+            user_model,
+            role_model,
+            webauthn_model,
         )
 
     def commit(self):
@@ -629,6 +848,7 @@ class MongoEngineUserDatastore(MongoEngineDatastore, UserDatastore):
     :param db:
     :param user_model: See :ref:`Models <models_topic>`.
     :param role_model: See :ref:`Models <models_topic>`.
+    :param webauthn_model: See :ref:`Models <models_topic>`.
     """
 
     def __init__(
@@ -636,9 +856,10 @@ class MongoEngineUserDatastore(MongoEngineDatastore, UserDatastore):
         db: "flask_mongoengine.MongoEngine",
         user_model: t.Type["User"],
         role_model: t.Type["Role"],
+        webauthn_model: t.Optional[t.Type["WebAuthn"]] = None,
     ):
         MongoEngineDatastore.__init__(self, db)
-        UserDatastore.__init__(self, user_model, role_model)
+        UserDatastore.__init__(self, user_model, role_model, webauthn_model)
 
     def find_user(self, case_insensitive=False, **kwargs):
         from mongoengine.queryset.visitor import Q, QCombination
@@ -653,17 +874,61 @@ class MongoEngineUserDatastore(MongoEngineDatastore, UserDatastore):
                     raise ValueError("Case insensitive option only supports single key")
                 attr, identifier = kwargs.popitem()
                 query = {f"{attr}__iexact": identifier}
-                return self.user_model.objects(**query).first()
+                obj = self.user_model.objects(**query).first()
             else:
                 queries = map(lambda i: Q(**{i[0]: i[1]}), kwargs.items())
                 query = QCombination(QCombination.AND, queries)
-                return self.user_model.objects(query).first()
+                obj = self.user_model.objects(query).first()
         except ValidationError:  # pragma: no cover
             return None
+        return obj
 
     def find_role(self, role):
         return self.role_model.objects(name=role).first()
 
+    def find_webauthn(self, credential_id: bytes) -> t.Union["WebAuthn", None]:
+        if not self.webauthn_model:
+            raise NotImplementedError
+
+        obj = self.webauthn_model.objects(  # type: ignore
+            credential_id=credential_id
+        ).first()
+        return obj
+
+    def create_webauthn(
+        self,
+        user: "User",
+        credential_id: bytes,
+        public_key: bytes,
+        name: str,
+        sign_count: int,
+        usage: str,
+        device_type: str,
+        backup_state: bool,
+        transports: t.Optional[t.List[str]] = None,
+        extensions: t.Optional[str] = None,
+        **kwargs: t.Any,
+    ) -> None:
+        if not hasattr(self, "webauthn_model") or not self.webauthn_model:
+            raise NotImplementedError
+        webauthn = self.webauthn_model(
+            user=user,
+            credential_id=credential_id,
+            public_key=public_key,
+            name=name,
+            sign_count=sign_count,
+            usage=usage,
+            device_type=device_type,
+            backup_state=backup_state,
+            transports=transports,
+            extensions=extensions,
+            lastuse_datetime=datetime.datetime.utcnow(),
+            **kwargs,
+        )
+        user.webauthn.append(webauthn)
+        self.put(webauthn)  # type: ignore
+        self.put(user)  # type: ignore
+
 
 class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
     """A UserDatastore implementation that assumes the
@@ -671,22 +936,19 @@ class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
     `Peewee Flask utils \
        <https://docs.peewee-orm.com/en/latest/peewee/playhouse.html#flask-utils>`_
     for datastore transactions.
-
-    :param db:
-    :param user_model: See :ref:`Models <models_topic>`.
-    :param role_model: See :ref:`Models <models_topic>`.
-    :param role_link:
     """
 
-    def __init__(self, db, user_model, role_model, role_link):
+    def __init__(self, db, user_model, role_model, role_link, webauthn_model=None):
         """
         :param db:
         :param user_model: A user model class definition
         :param role_model: A role model class definition
         :param role_link: A model implementing the many-to-many user-role relation
+        :param webauthn_model: A webauthn model class definition
+
         """
         PeeweeDatastore.__init__(self, db)
-        UserDatastore.__init__(self, user_model, role_model)
+        UserDatastore.__init__(self, user_model, role_model, webauthn_model)
         self.UserRole = role_link
 
     def find_user(self, case_insensitive=False, **kwargs):
@@ -760,6 +1022,46 @@ class PeeweeUserDatastore(PeeweeDatastore, UserDatastore):
         else:
             return False
 
+    def find_webauthn(self, credential_id):
+        if not self.webauthn_model:
+            raise NotImplementedError
+        try:
+            return self.webauthn_model.filter(credential_id=credential_id).get()
+        except self.webauthn_model.DoesNotExist:
+            return None
+
+    def create_webauthn(
+        self,
+        user: "User",
+        credential_id: bytes,
+        public_key: bytes,
+        name: str,
+        sign_count: int,
+        usage: str,
+        device_type: str,
+        backup_state: bool,
+        transports: t.Optional[t.List[str]] = None,
+        extensions: t.Optional[str] = None,
+        **kwargs: t.Any,
+    ) -> None:
+        if not hasattr(self, "webauthn_model") or not self.webauthn_model:
+            raise NotImplementedError
+        webauthn = self.webauthn_model(
+            user=user,
+            credential_id=credential_id,
+            public_key=public_key,
+            name=name,
+            sign_count=sign_count,
+            usage=usage,
+            device_type=device_type,
+            backup_state=backup_state,
+            transports=transports,
+            extensions=extensions,
+            lastuse_datetime=datetime.datetime.utcnow(),
+            **kwargs,
+        )
+        self.put(webauthn)  # type: ignore
+
 
 class PonyUserDatastore(PonyDatastore, UserDatastore):
     """A UserDatastore implementation that assumes the
@@ -773,11 +1075,12 @@ class PonyUserDatastore(PonyDatastore, UserDatastore):
     :param db:
     :param user_model: See :ref:`Models <models_topic>`.
     :param role_model: See :ref:`Models <models_topic>`.
+    :param webauthn_model: See :ref:`Models <models_topic>`.
     """
 
-    def __init__(self, db, user_model, role_model):
+    def __init__(self, db, user_model, role_model, webauthn_model=None):
         PonyDatastore.__init__(self, db)
-        UserDatastore.__init__(self, user_model, role_model)
+        UserDatastore.__init__(self, user_model, role_model, webauthn_model)
 
     @with_pony_session
     def find_user(self, case_insensitive=False, **kwargs):
@@ -812,7 +1115,7 @@ if t.TYPE_CHECKING:  # pragma: no cover
     # Normally - the application creates the Models and glues them together
     # For typing we do that here since we don't know which DB interface they
     # will pick.
-    from .core import UserMixin, RoleMixin
+    from .core import UserMixin, RoleMixin, WebAuthnMixin
 
     class CanonicalUserDatastore(Datastore, UserDatastore):
         pass
@@ -821,11 +1124,12 @@ if t.TYPE_CHECKING:  # pragma: no cover
         id: int
         email: str
         username: t.Optional[str]
-        password: str
+        password: t.Optional[str]
         active: bool
         fs_uniquifier: str
         fs_token_uniquifier: str
-        confirmed_at: datetime.datetime
+        fs_webauthn_user_handle: str
+        confirmed_at: t.Optional[datetime.datetime]
         last_login_at: datetime.datetime
         current_login_at: datetime.datetime
         last_login_ip: t.Optional[str]
@@ -834,11 +1138,13 @@ if t.TYPE_CHECKING:  # pragma: no cover
         tf_primary_method: t.Optional[str]
         tf_totp_secret: t.Optional[str]
         tf_phone_number: t.Optional[str]
+        mf_recovery_codes: t.Optional[t.List[str]]
         us_phone_number: t.Optional[str]
         us_totp_secrets: t.Optional[t.Union[str, bytes]]
         create_datetime: datetime.datetime
         update_datetime: datetime.datetime
         roles: t.List["Role"]
+        webauthn: t.List["WebAuthn"]
 
         def __init__(self, **kwargs):
             ...
@@ -847,8 +1153,25 @@ if t.TYPE_CHECKING:  # pragma: no cover
         id: int
         name: str
         description: t.Optional[str]
-        permissions: t.Optional[t.Union[str, set, list]]
+        permissions: t.Optional[t.List[str]]
         update_datetime: datetime.datetime
 
         def __init__(self, **kwargs):
             ...
+
+    class WebAuthn(WebAuthnMixin):
+        id: int
+        name: str
+        credential_id: bytes
+        public_key: bytes
+        sign_count: int
+        transports: t.Optional[t.List[str]]
+        backup_state: bool
+        device_type: str
+        extensions: t.Optional[str]
+        lastuse_datetime: datetime.datetime
+        user_id: int
+        usage: str
+
+        def __init__(self, **kwargs):
+            ...
