@@ -5,6 +5,7 @@
     Test fixtures and what not
 
     :copyright: (c) 2017 by CERN.
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -16,11 +17,9 @@ from datetime import datetime
 from urllib.parse import urlsplit
 
 import pytest
-from flask import Flask, Response, render_template
-from flask import jsonify
+from flask import Flask, Response, jsonify, render_template
 from flask import request as flask_request
-from flask.json import JSONEncoder
-from flask_mail import Mail
+from flask_mailman import Mail
 
 from flask_security import (
     MongoEngineUserDatastore,
@@ -31,6 +30,7 @@ from flask_security import (
     SQLAlchemySessionUserDatastore,
     SQLAlchemyUserDatastore,
     UserMixin,
+    WebAuthnMixin,
     auth_required,
     auth_token_required,
     http_auth_required,
@@ -59,9 +59,14 @@ if t.TYPE_CHECKING:  # pragma: no cover
     from flask.testing import FlaskClient
 
 
+class SecurityFixture(Flask):
+    security: Security
+    mail: Mail
+
+
 @pytest.fixture()
-def app(request: pytest.FixtureRequest) -> "Flask":
-    app = Flask(__name__)
+def app(request: pytest.FixtureRequest) -> "SecurityFixture":
+    app = SecurityFixture(__name__)
     app.response_class = Response
     app.debug = True
     app.config["SECRET_KEY"] = "secret"
@@ -92,14 +97,18 @@ def app(request: pytest.FixtureRequest) -> "Flask":
         "confirmable",
         "two_factor",
         "unified_signin",
+        "webauthn",
     ]:
         app.config["SECURITY_" + opt.upper()] = opt in request.keywords
 
-    pytest_major = int(pytest.__version__.split(".")[0])
-    if pytest_major >= 4:
-        marker_getter = request.node.get_closest_marker
-    else:
-        marker_getter = request.keywords.get
+    marker_getter = request.node.get_closest_marker
+
+    # Import webauthn, or skip test if webauthn isn't installed
+    webauthn_test = marker_getter("webauthn")
+    if webauthn_test is not None:
+        pytest.importorskip("webauthn")
+
+    # Override config settings as requested for this test
     settings = marker_getter("settings")
     if settings is not None:
         for key, value in settings.kwargs.items():
@@ -110,7 +119,6 @@ def app(request: pytest.FixtureRequest) -> "Flask":
             app.config[key.upper()] = value
 
     app.mail = Mail(app)  # type: ignore
-    app.json_encoder = JSONEncoder  # type: ignore
 
     # use babel marker to signify tests that need babel extension.
     babel = marker_getter("babel")
@@ -224,7 +232,7 @@ def app(request: pytest.FixtureRequest) -> "Flask":
         return render_template("index.html", content="Unauthorized")
 
     @app.route("/fresh", methods=["GET", "POST"])
-    @auth_required(within=0)
+    @auth_required(within=60)
     def fresh():
         if app.security._want_json(flask_request):
             return jsonify(title="Fresh Only")
@@ -234,8 +242,8 @@ def app(request: pytest.FixtureRequest) -> "Flask":
     def revert_forms():
         # Some forms/tests have dynamic fields - be sure to revert them.
         if hasattr(app, "security"):
-            if hasattr(app.security.login_form, "email"):
-                del app.security.login_form.email
+            if hasattr(app.security.login_form, "username"):
+                del app.security.login_form.username
             if hasattr(app.security.register_form, "username"):
                 del app.security.register_form.username
             if hasattr(app.security.confirm_register_form, "username"):
@@ -246,14 +254,16 @@ def app(request: pytest.FixtureRequest) -> "Flask":
 
 
 @pytest.fixture()
-def mongoengine_datastore(request, app, tmpdir, realdburl):
-    return mongoengine_setup(request, app, tmpdir, realdburl)
+def mongoengine_datastore(request, app, tmpdir, realmongodburl):
+    return mongoengine_setup(request, app, tmpdir, realmongodburl)
 
 
-def mongoengine_setup(request, app, tmpdir, realdburl):
+def mongoengine_setup(request, app, tmpdir, realmongodburl):
+    # To run against a realdb: mongod --dbpath <somewhere>
     pytest.importorskip("flask_mongoengine")
     from flask_mongoengine import MongoEngine
     from mongoengine.fields import (
+        BinaryField,
         BooleanField,
         DateTimeField,
         IntField,
@@ -261,11 +271,12 @@ def mongoengine_setup(request, app, tmpdir, realdburl):
         ReferenceField,
         StringField,
     )
+    from mongoengine import PULL, CASCADE, disconnect_all
 
-    db_name = "flask_security_test_%s" % str(time.time()).replace(".", "_")
+    db_name = "flask_security_test"
     app.config["MONGODB_SETTINGS"] = {
         "db": db_name,
-        "host": "mongomock://localhost",
+        "host": realmongodburl if realmongodburl else "mongomock://localhost",
         "port": 27017,
         "alias": db_name,
     }
@@ -275,12 +286,38 @@ def mongoengine_setup(request, app, tmpdir, realdburl):
     class Role(db.Document, RoleMixin):
         name = StringField(required=True, unique=True, max_length=80)
         description = StringField(max_length=255)
-        permissions = StringField(max_length=255)
+        permissions = ListField(required=False)
         meta = {"db_alias": db_name}
 
+    class WebAuthn(db.Document, WebAuthnMixin):
+        credential_id = BinaryField(primary_key=True, max_bytes=1024, required=True)
+        public_key = BinaryField(required=True)
+        sign_count = IntField(default=0)
+        transports = ListField(required=False)
+        backup_state = BooleanField(required=True)
+        device_type = StringField(max_length=64, required=True)
+
+        # a JSON string as returned from registration
+        extensions = StringField(max_length=255)
+        lastuse_datetime = DateTimeField(required=True)
+        # name is provided by user - we make sure it is unique per user
+        name = StringField(max_length=64, required=True)
+        usage = StringField(max_length=64, required=True)
+        # we need to be able to look up a user from a credential_id
+        user = ReferenceField("User")
+        # user_id = ObjectIdField(required=True)
+        meta = {"db_alias": db_name}
+
+        def get_user_mapping(self) -> t.Dict[str, str]:
+            """
+            Return the mapping from webauthn back to User
+            """
+            return dict(id=self.user.id)
+
     class User(db.Document, UserMixin):
         email = StringField(unique=True, max_length=255)
         fs_uniquifier = StringField(unique=True, max_length=64, required=True)
+        fs_webauthn_user_handle = StringField(unique=True, max_length=64)
         username = StringField(unique=True, required=False, sparse=True, max_length=255)
         password = StringField(required=False, max_length=255)
         security_number = IntField(unique=True, required=False, sparse=True)
@@ -289,25 +326,38 @@ def mongoengine_setup(request, app, tmpdir, realdburl):
         tf_primary_method = StringField(max_length=255)
         tf_totp_secret = StringField(max_length=255)
         tf_phone_number = StringField(max_length=255)
+        mf_recovery_codes = ListField(required=False)
         us_totp_secrets = StringField()
-        us_phone_number = StringField(max_length=255)
+        us_phone_number = StringField(
+            max_length=255, unique=True, required=False, sparse=True
+        )
         last_login_ip = StringField(max_length=100)
         current_login_ip = StringField(max_length=100)
         login_count = IntField()
         active = BooleanField(default=True)
         confirmed_at = DateTimeField()
         roles = ListField(ReferenceField(Role), default=[])
+        webauthn = ListField(
+            ReferenceField(WebAuthn, reverse_delete_rule=PULL), default=[]
+        )
         meta = {"db_alias": db_name}
 
+        def get_security_payload(self):
+            return {"email": str(self.email)}
+
+    db.Document.register_delete_rule(WebAuthn, "user", CASCADE)
+
     def tear_down():
         with app.app_context():
             User.drop_collection()
             Role.drop_collection()
+            WebAuthn.drop_collection()
             db.connection.drop_database(db_name)
+            disconnect_all()
 
     request.addfinalizer(tear_down)
 
-    return MongoEngineUserDatastore(db, User, Role)
+    return MongoEngineUserDatastore(db, User, Role, WebAuthn)
 
 
 @pytest.fixture()
@@ -318,7 +368,8 @@ def sqlalchemy_datastore(request, app, tmpdir, realdburl):
 def sqlalchemy_setup(request, app, tmpdir, realdburl):
     pytest.importorskip("flask_sqlalchemy")
     from flask_sqlalchemy import SQLAlchemy
-    from flask_security.models import fsqla_v2 as fsqla
+    from sqlalchemy import Column, Integer
+    from flask_security.models import fsqla_v3 as fsqla
 
     if realdburl:
         db_url, db_info = _setup_realdb(realdburl)
@@ -326,10 +377,6 @@ def sqlalchemy_setup(request, app, tmpdir, realdburl):
     else:
         app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
 
-    # In Flask-SQLAlchemy >= 3.0.0 queries are no longer logged automatically,
-    # even in debug or testing mode.
-    app.config["SQLALCHEMY_RECORD_QUERIES"] = True
-
     db = SQLAlchemy(app)
 
     fsqla.FsModels.set_db_info(db)
@@ -337,13 +384,15 @@ def sqlalchemy_setup(request, app, tmpdir, realdburl):
     class Role(db.Model, fsqla.FsRoleMixin):
         pass
 
+    class WebAuthn(db.Model, fsqla.FsWebAuthnMixin):
+        pass
+
     class User(db.Model, fsqla.FsUserMixin):
-        security_number = db.Column(db.Integer, unique=True)
-        # For testing allow null passwords.
-        password = db.Column(db.String(255), nullable=True)
+        security_number = Column(Integer, unique=True)
 
         def get_security_payload(self):
-            # Make sure we still properly hook up to flask JSONEncoder
+            # Make sure we still properly hook up to flask's JSON extension
+            # which handles datetime
             return {"email": str(self.email), "last_update": self.update_datetime}
 
     with app.app_context():
@@ -351,13 +400,12 @@ def sqlalchemy_setup(request, app, tmpdir, realdburl):
 
     def tear_down():
         if realdburl:
-            with app.app_context():
-                db.drop_all()
-                _teardown_realdb(db_info)
+            db.drop_all()
+            _teardown_realdb(db_info)
 
     request.addfinalizer(tear_down)
 
-    return SQLAlchemyUserDatastore(db, User, Role)
+    return SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
 
 
 @pytest.fixture()
@@ -366,21 +414,26 @@ def sqlalchemy_session_datastore(request, app, tmpdir, realdburl):
 
 
 def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
+    """
+    Note that we test having a different user id column name here.
+    """
     pytest.importorskip("sqlalchemy")
     from sqlalchemy import create_engine
     from sqlalchemy.orm import scoped_session, sessionmaker, relationship, backref
-    from sqlalchemy.ext.declarative import declarative_base
+    from sqlalchemy.ext.declarative import declarative_base, declared_attr
+    from sqlalchemy.ext.mutable import MutableList
     from sqlalchemy.sql import func
     from sqlalchemy import (
         Boolean,
         DateTime,
         Column,
         Integer,
+        LargeBinary,
         String,
         Text,
         ForeignKey,
-        UnicodeText,
     )
+    from flask_security.models.fsqla_v3 import AsaList
 
     f, path = tempfile.mkstemp(
         prefix="flask-security-test-db", suffix=".db", dir=str(tmpdir)
@@ -395,6 +448,44 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
     Base = declarative_base()
     Base.query = db_session.query_property()
 
+    class WebAuthn(Base, WebAuthnMixin):
+        __tablename__ = "webauthn"
+
+        id = Column(Integer, primary_key=True)
+        credential_id = Column(
+            LargeBinary(1024), index=True, unique=True, nullable=False
+        )
+        public_key = Column(LargeBinary, nullable=False)
+        sign_count = Column(Integer, default=0)
+        transports = Column(MutableList.as_mutable(AsaList()), nullable=True)
+
+        # a JSON string as returned from registration
+        extensions = Column(String(255), nullable=True)
+        create_datetime = Column(
+            type_=DateTime, nullable=False, server_default=func.now()
+        )
+        lastuse_datetime = Column(type_=DateTime, nullable=False)
+        # name is provided by user - we make sure is unique per user
+        name = Column(String(64), nullable=False)
+        usage = Column(String(64), nullable=False)
+        backup_state = Column(Boolean, nullable=False)  # Upcoming post V3 spec
+        device_type = Column(String(64), nullable=False)
+
+        @declared_attr
+        def myuser_id(cls):
+            return Column(
+                Integer,
+                ForeignKey("user.myuserid", ondelete="CASCADE"),
+                nullable=False,
+            )
+
+        def get_user_mapping(self) -> t.Dict[str, t.Any]:
+            """
+            Return the filter needed by find_user() to get the user
+            associated with this webauthn credential.
+            """
+            return dict(myuserid=self.myuser_id)
+
     class RolesUsers(Base):
         __tablename__ = "roles_users"
         id = Column(Integer(), primary_key=True)
@@ -406,7 +497,7 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
         myroleid = Column(Integer(), primary_key=True)
         name = Column(String(80), unique=True)
         description = Column(String(255))
-        permissions = Column(UnicodeText, nullable=True)
+        permissions = Column(MutableList.as_mutable(AsaList()), nullable=True)
         update_datetime = Column(
             DateTime,
             nullable=False,
@@ -418,6 +509,7 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
         __tablename__ = "user"
         myuserid = Column(Integer, primary_key=True)
         fs_uniquifier = Column(String(64), unique=True, nullable=False)
+        fs_webauthn_user_handle = Column(String(64), unique=True, nullable=True)
         email = Column(String(255), unique=True)
         username = Column(String(255), unique=True, nullable=True)
         password = Column(String(255))
@@ -427,8 +519,9 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
         tf_primary_method = Column(String(255), nullable=True)
         tf_totp_secret = Column(String(255), nullable=True)
         tf_phone_number = Column(String(255), nullable=True)
+        mf_recovery_codes = Column(MutableList.as_mutable(AsaList()), nullable=True)
         us_totp_secrets = Column(Text, nullable=True)
-        us_phone_number = Column(String(64), nullable=True)
+        us_phone_number = Column(String(64), nullable=True, unique=True)
         last_login_ip = Column(String(100))
         current_login_ip = Column(String(100))
         login_count = Column(Integer)
@@ -444,8 +537,13 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
             onupdate=datetime.utcnow,
         )
 
+        @declared_attr
+        def webauthn(cls):
+            return relationship("WebAuthn", backref="users", cascade="all, delete")
+
         def get_security_payload(self):
-            # Make sure we still properly hook up to flask JSONEncoder
+            # Make sure we still properly hook up to flask's JSON extension
+            # which handles datetime
             return {"email": str(self.email), "last_update": self.update_datetime}
 
     with app.app_context():
@@ -458,7 +556,7 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
 
     request.addfinalizer(tear_down)
 
-    return SQLAlchemySessionUserDatastore(db_session, User, Role)
+    return SQLAlchemySessionUserDatastore(db_session, User, Role, WebAuthn)
 
 
 @pytest.fixture()
@@ -471,8 +569,10 @@ def peewee_setup(request, app, tmpdir, realdburl):
     from peewee import (
         TextField,
         DateTimeField,
+        Field,
         IntegerField,
         BooleanField,
+        BlobField,
         ForeignKeyField,
         CharField,
     )
@@ -503,14 +603,29 @@ def peewee_setup(request, app, tmpdir, realdburl):
 
     db = FlaskDB(app)
 
+    class AsaList(Field):
+        field_type = "text"
+
+        def db_value(self, value):
+            try:
+                return ",".join(value)
+            except TypeError:
+                return value
+
+        def python_value(self, value):
+            if value:
+                return value.split(",")
+            return []
+
     class Role(RoleMixin, db.Model):
         name = CharField(unique=True, max_length=80)
         description = TextField(null=True)
-        permissions = TextField(null=True)
+        permissions = AsaList(null=True)
 
     class User(UserMixin, db.Model):
         email = TextField(unique=True, null=False)
         fs_uniquifier = TextField(unique=True, null=False)
+        fs_webauthn_user_handle = TextField(unique=True, null=True)
         username = TextField(unique=True, null=True)
         security_number = IntegerField(null=True)
         password = TextField(null=True)
@@ -519,14 +634,36 @@ def peewee_setup(request, app, tmpdir, realdburl):
         tf_primary_method = TextField(null=True)
         tf_totp_secret = TextField(null=True)
         tf_phone_number = TextField(null=True)
+        mf_recovery_codes = AsaList(null=True)
         us_totp_secrets = TextField(null=True)
-        us_phone_number = TextField(null=True)
+        us_phone_number = TextField(null=True, unique=True)
         last_login_ip = TextField(null=True)
         current_login_ip = TextField(null=True)
         login_count = IntegerField(null=True)
         active = BooleanField(default=True)
         confirmed_at = DateTimeField(null=True)
 
+        def get_security_payload(self):
+            return {"email": str(self.email)}
+
+    class WebAuthn(WebAuthnMixin, db.Model):
+        credential_id = BlobField(unique=True, null=False, index=True)
+        public_key = BlobField(null=False)
+        sign_count = IntegerField(default=0)
+        transports = AsaList(null=True)
+
+        # a JSON string as returned from registration
+        extensions = TextField(null=True)
+        lastuse_datetime = DateTimeField(null=False)
+        # name is provided by user - we make sure is unique per user
+        name = TextField(null=False)
+        usage = TextField(null=False)
+        backup_state = BooleanField()
+        device_type = TextField(null=False)
+
+        # This creates a real column called user_id
+        user = ForeignKeyField(User, backref="webauthn")
+
     class UserRoles(db.Model):
         """Peewee does not have built-in many-to-many support, so we have to
         create this mapping class to link users to roles."""
@@ -540,7 +677,7 @@ def peewee_setup(request, app, tmpdir, realdburl):
             return self.role.get_permissions()
 
     with app.app_context():
-        for Model in (Role, User, UserRoles):
+        for Model in (Role, User, UserRoles, WebAuthn):
             Model.drop_table()
             Model.create_table()
 
@@ -555,7 +692,7 @@ def peewee_setup(request, app, tmpdir, realdburl):
 
     request.addfinalizer(tear_down)
 
-    return PeeweeUserDatastore(db, User, Role, UserRoles)
+    return PeeweeUserDatastore(db, User, Role, UserRoles, WebAuthn)
 
 
 @pytest.fixture()
@@ -564,6 +701,7 @@ def pony_datastore(request, app, tmpdir, realdburl):
 
 
 def pony_setup(request, app, tmpdir, realdburl):
+
     pytest.importorskip("pony")
     from pony.orm import Database, Optional, Required, Set
     from pony.orm.core import SetInstance
@@ -629,11 +767,11 @@ def pony_setup(request, app, tmpdir, realdburl):
 
 @pytest.fixture()
 def sqlalchemy_app(
-    app: "Flask", sqlalchemy_datastore: SQLAlchemyUserDatastore
-) -> t.Callable[[], "Flask"]:
-    def create() -> "Flask":
+    app: SecurityFixture, sqlalchemy_datastore: SQLAlchemyUserDatastore
+) -> t.Callable[[], SecurityFixture]:
+    def create() -> SecurityFixture:
         security = Security(app, datastore=sqlalchemy_datastore)
-        app.security = security  # type: ignore
+        app.security = security
         return app
 
     return create
@@ -692,13 +830,13 @@ def client_nc(request, sqlalchemy_app):
 
 
 @pytest.fixture(params=["cl-sqlalchemy", "c2", "cl-mongo", "cl-peewee"])
-def clients(request, app, tmpdir, realdburl):
+def clients(request, app, tmpdir, realdburl, realmongodburl):
     if request.param == "cl-sqlalchemy":
         ds = sqlalchemy_setup(request, app, tmpdir, realdburl)
     elif request.param == "c2":
         ds = sqlalchemy_session_setup(request, app, tmpdir, realdburl)
     elif request.param == "cl-mongo":
-        ds = mongoengine_setup(request, app, tmpdir, realdburl)
+        ds = mongoengine_setup(request, app, tmpdir, realmongodburl)
     elif request.param == "cl-peewee":
         ds = peewee_setup(request, app, tmpdir, realdburl)
     elif request.param == "cl-pony":
@@ -739,13 +877,13 @@ def get_message_local(app):
 @pytest.fixture(
     params=["sqlalchemy", "sqlalchemy-session", "mongoengine", "peewee", "pony"]
 )
-def datastore(request, app, tmpdir, realdburl):
+def datastore(request, app, tmpdir, realdburl, realmongodburl):
     if request.param == "sqlalchemy":
         rv = sqlalchemy_setup(request, app, tmpdir, realdburl)
     elif request.param == "sqlalchemy-session":
         rv = sqlalchemy_session_setup(request, app, tmpdir, realdburl)
     elif request.param == "mongoengine":
-        rv = mongoengine_setup(request, app, tmpdir, realdburl)
+        rv = mongoengine_setup(request, app, tmpdir, realmongodburl)
     elif request.param == "peewee":
         rv = peewee_setup(request, app, tmpdir, realdburl)
     elif request.param == "pony":
@@ -779,6 +917,13 @@ def pytest_addoption(parser):
         help="""Set url for using real database for testing.
         For postgres: 'postgresql://user:password@host/')""",
     )
+    parser.addoption(
+        "--realmongodburl",
+        action="store",
+        default=None,
+        help="""Set url for using real mongo database for testing.
+        e.g. 'localhost'""",
+    )
 
 
 @pytest.fixture(scope="session")
@@ -796,6 +941,16 @@ def realdburl(request):
     return request.config.option.realdburl
 
 
+@pytest.fixture(scope="session")
+def realmongodburl(request):
+    """
+    Support running datastore tests against a real Mongo DB.
+    --realmongodburl "localhost"
+
+    """
+    return request.config.option.realmongodburl
+
+
 def _setup_realdb(realdburl):
     """
     Called when we want to run unit tests against a real DB.
