@@ -32,7 +32,9 @@ from tests.test_utils import (
     json_authenticate,
     logout,
     populate_data,
+    reset_fresh,
 )
+from tests.test_webauthn import HackWebauthnUtil, reg_2_keys
 
 from flask import Flask, abort, request, Response
 from flask_security import Security
@@ -55,11 +57,13 @@ from flask_security.forms import (
 )
 from flask_security import auth_required, roles_required
 from flask_security.utils import (
+    base_render_json,
     encode_string,
     json_error_response,
     get_request_attr,
     hash_data,
     send_mail,
+    uia_email_mapper,
     uia_phone_mapper,
     validate_redirect_url,
     verify_hash,
@@ -93,7 +97,9 @@ def test_my_mail_util(app, sqlalchemy_datastore):
 
 
 def test_register_blueprint_flag(app, sqlalchemy_datastore):
-    app.security = Security(app, datastore=Security, register_blueprint=False)
+    app.security = Security(
+        app, datastore=sqlalchemy_datastore, register_blueprint=False
+    )
     client = app.test_client()
     response = client.get("/login")
     assert response.status_code == 404
@@ -102,9 +108,15 @@ def test_register_blueprint_flag(app, sqlalchemy_datastore):
 @pytest.mark.registerable()
 @pytest.mark.recoverable()
 @pytest.mark.changeable()
+@pytest.mark.settings(
+    USER_IDENTITY_ATTRIBUTES=[
+        {"email": {"mapper": uia_email_mapper}},
+        {"username": {"mapper": lambda x: x}},
+    ]
+)
 def test_basic_custom_forms(app, sqlalchemy_datastore):
     class MyLoginForm(LoginForm):
-        email = EmailField("My Login Email Address Field")
+        username = StringField("My Login Username Field")
 
     class MyRegisterForm(RegisterForm):
         email = EmailField("My Register Email Address Field")
@@ -135,7 +147,7 @@ def test_basic_custom_forms(app, sqlalchemy_datastore):
     client = app.test_client()
 
     response = client.get("/login")
-    assert b"My Login Email Address Field" in response.data
+    assert b"My Login Username Field" in response.data
 
     response = client.get("/register")
     assert b"My Register Email Address Field" in response.data
@@ -342,6 +354,9 @@ def test_password_unicode_password_salt(client):
     assert b"Welcome matt@lp.com" in response.data
 
 
+@pytest.mark.filterwarnings(
+    "ignore:.*'unauthorized_handler' has been replaced.*:DeprecationWarning"
+)
 def test_set_unauthorized_handler(app, client):
     @app.security.unauthorized_handler
     def unauthorized():
@@ -421,9 +436,9 @@ def test_form_required_local_message(app, sqlalchemy_datastore):
     assert b"Key" not in response.data
 
 
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_without_babel(app, client):
     # Test if babel modules exist but we don't init babel - things still work
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     response = client.get("/login")
     assert response.status_code == 200
 
@@ -443,15 +458,17 @@ def test_no_email_sender(app, sqlalchemy_datastore):
 
     with app.app_context():
         user = TestUser("matt@lp.com")
-        with app.mail.record_messages() as outbox:
-            send_mail("Test Default Sender", user.email, "welcome", user=user)
-            assert 1 == len(outbox)
-            assert "test@testme.com" == outbox[0].sender
+        send_mail("Test Default Sender", user.email, "welcome", user=user)
+        outbox = app.mail.outbox
+        assert 1 == len(outbox)
+        assert "test@testme.com" == outbox[0].from_email
 
 
 def test_sender_tuple(app, sqlalchemy_datastore):
     """Verify that if sender is a (name, address) tuple,
     in the received email sender is properly formatted as "name <address>"
+    Flask-Mail takes tuples - Flask-Mailman takes them - however the
+    local-mem backend doesn't format them correctly (SMTP backend doesn't work either?)
     """
     app.config["MAIL_DEFAULT_SENDER"] = ("Test User", "test@testme.com")
 
@@ -464,15 +481,15 @@ def test_sender_tuple(app, sqlalchemy_datastore):
 
     with app.app_context():
         user = TestUser("matt@lp.com")
-        with app.mail.record_messages() as outbox:
-            send_mail("Test Tuple Sender", user.email, "welcome", user=user)
-            assert 1 == len(outbox)
-            assert "Test User <test@testme.com>" == outbox[0].sender
+        send_mail("Test Tuple Sender", user.email, "welcome", user=user)
+        outbox = app.mail.outbox
+        assert 1 == len(outbox)
+        assert outbox[0].from_email == "Test User <test@testme.com>"
 
 
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 @pytest.mark.babel()
 def test_xlation(app, client):
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     response = client.get("/login")
@@ -484,7 +501,6 @@ def test_xlation(app, client):
 
 
 @pytest.mark.babel()
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
     # Test changing a single MSG and having an additional translation dir
     # Flask-BabelEx doesn't support lists of directories..
@@ -503,6 +519,7 @@ def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
         app, sqlalchemy_datastore, **{"SECURITY_I18N_DIRNAME": i18n_dirname}
     )
 
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     app.config["SECURITY_MSG_INVALID_PASSWORD"] = ("Password no-worky", "error")
@@ -513,8 +530,8 @@ def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
 
 
 @pytest.mark.babel()
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_form_labels(app, sqlalchemy_datastore):
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
     app.security.init_app(app, sqlalchemy_datastore)
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
@@ -540,7 +557,6 @@ def test_form_labels(app, sqlalchemy_datastore):
 
 
 @pytest.mark.babel()
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_wtform_xlation(app, sqlalchemy_datastore):
     # Make sure wtform xlations work
     class MyLoginForm(LoginForm):
@@ -548,6 +564,7 @@ def test_wtform_xlation(app, sqlalchemy_datastore):
             "FixedLength", validators=[DataRequired(), Length(3, 3)]
         )
 
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
     app.security.init_app(app, datastore=sqlalchemy_datastore, login_form=MyLoginForm)
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
@@ -562,7 +579,7 @@ def test_wtform_xlation(app, sqlalchemy_datastore):
         "/login", json=data, headers={"Content-Type": "application/json"}
     )
     assert response.status_code == 400
-    flerror = response.json["response"]["errors"]["fixed_length"][0]
+    flerror = response.json["response"]["field_errors"]["fixed_length"][0]
     # This is completely dependent on WTforms translations....
     assert (
         flerror == "Le doit contenir exactement 3 caractères."
@@ -577,6 +594,7 @@ def test_per_request_xlate(app, client):
 
     babel = app.extensions["babel"]
 
+    @babel.localeselector
     def get_locale():
         # For a given session - set lang based on first request.
         # Honor explicit url request first
@@ -588,9 +606,6 @@ def test_per_request_xlate(app, client):
                 session["lang"] = locale
         return session.get("lang", None).replace("-", "_")
 
-    babel.locale_selector_func = get_locale
-    babel.locale_selector = get_locale  # Flask-Babel >= 3.0.0
-
     response = client.get("/login", headers=[("Accept-Language", "fr")])
     assert b'<label for="password">Mot de passe</label>' in response.data
     # make sure template contents get xlated (not just form).
@@ -623,7 +638,7 @@ def test_per_request_xlate(app, client):
         headers={"Content-Type": "application/json"},
     )
     assert response.status_code == 400
-    assert response.json["response"]["errors"]["new_password"] == [
+    assert response.json["response"]["field_errors"]["new_password"] == [
         "Merci d'indiquer un mot de passe"
     ]
 
@@ -650,6 +665,7 @@ def test_zxcvbn_xlate(app):
 @pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
 @pytest.mark.settings(password_check_breached="strict")
 def test_breached(app, sqlalchemy_datastore):
+
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
@@ -678,6 +694,7 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
     password_complexity_checker="zxcvbn",
 )
 def test_breached_cnt(app, sqlalchemy_datastore):
+
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
@@ -717,21 +734,22 @@ def test_json_error_response_string():
     """Unit test for correct response when a string is given."""
     error_msg = "This is an error!"
     response = json_error_response(errors=error_msg)
-    assert "error" in response
-    assert "errors" not in response
-    assert response["error"] == error_msg
+    assert "field_errors" not in response
+    assert response["errors"][0] == error_msg
 
 
 def test_json_error_response_dict():
     """Unit test for correct response when a dict is given."""
     error_msg = {
-        "e-mail": "The e-mail address is already in the system.",
-        "name": "The name is too long.",
+        "e-mail": ["The e-mail address is already in the system."],
+        "name": ["The name is too long.", "Nice name"],
     }
-    response = json_error_response(errors=error_msg)
+    all_msgs = []
+    [all_msgs.extend(m) for m in error_msg.values()]
+    response = json_error_response(field_errors=error_msg)
     assert "errors" in response
-    assert "error" not in response
-    assert response["errors"] == error_msg
+    assert "field_errors" in response
+    assert all(m in response["errors"] for m in all_msgs)
 
 
 def test_json_error_response_typeerror():
@@ -741,6 +759,16 @@ def test_json_error_response_typeerror():
         json_error_response(errors=error_msg)
 
 
+def test_json_form_errors(app, client):
+    """Test wtforms form level errors are correctly sent via json"""
+    with app.test_request_context():
+        form = ChangePasswordForm()
+        form.form_errors.append("I am an error")
+        response = base_render_json(form)
+        assert len(response.json["response"]["errors"]) == 1
+        assert response.json["response"]["errors"][0] == "I am an error"
+
+
 def test_method_view(app, client):
     # auth_required with flask method view
     from flask.views import MethodView
@@ -822,7 +850,7 @@ def test_authn_freshness(
     # Test json error response
     response = client.get("/myspecialview", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json and response.json["response"]["error"].encode(
+    assert response.json and response.json["response"]["errors"][0].encode(
         "utf-8"
     ) == get_message("REAUTHENTICATION_REQUIRED")
 
@@ -862,7 +890,7 @@ def test_authn_freshness_handler(app, client, get_message):
     # Test json error response
     response = client.get("/myspecialview", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json["response"]["error"] == "Oh No"
+    assert response.json["response"]["errors"][0] == "Oh No"
 
 
 def test_authn_freshness_callable(app, client, get_message):
@@ -882,7 +910,7 @@ def test_authn_freshness_callable(app, client, get_message):
 def test_default_authn_bp(app, client):
     """Test default reauthn handler with blueprint prefix"""
 
-    @auth_required(within=0.001, grace=0)
+    @auth_required(within=1, grace=0)
     def myview():
         return Response(status=200)
 
@@ -890,7 +918,7 @@ def test_default_authn_bp(app, client):
     authenticate(client, endpoint="/myprefix/login")
 
     # This should require additional authn and redirect to verify
-    time.sleep(0.1)
+    reset_fresh(client, within=timedelta(minutes=1))
     response = client.get("/myview", follow_redirects=False)
     assert response.status_code == 302
     assert "/myprefix/verify?next=http%3A%2F%2Flocalhost%2Fmyview" in response.location
@@ -941,10 +969,11 @@ def test_authn_freshness_nc(app, client_nc, get_message):
 def test_verify_fresh(app, client, get_message):
     # Hit a fresh-required endpoint and walk through verify
     authenticate(client)
+    reset_fresh(client, app.config["SECURITY_FRESHNESS"])
 
     with capture_flashes() as flashes:
         response = client.get("/fresh", follow_redirects=True)
-        assert b"Please Enter Your Password" in response.data
+        assert b"Please Reauthenticate" in response.data
     assert flashes[0]["category"] == "error"
     assert flashes[0]["message"].encode("utf-8") == get_message(
         "REAUTHENTICATION_REQUIRED"
@@ -955,13 +984,14 @@ def test_verify_fresh(app, client, get_message):
     )
     verify_url = matcher.group(1)
 
+    reset_fresh(client, app.config["SECURITY_FRESHNESS"])
     response = client.get(verify_url)
-    assert b"Please Enter Your Password" in response.data
+    assert b"Please Reauthenticate" in response.data
 
     response = client.post(
         verify_url, data=dict(password="not my password"), follow_redirects=False
     )
-    assert b"Please Enter Your Password" in response.data
+    assert b"Please Reauthenticate" in response.data
 
     response = client.post(
         verify_url, data=dict(password="password"), follow_redirects=False
@@ -978,12 +1008,13 @@ def test_verify_fresh_json(app, client, get_message):
     authenticate(client)
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
 
+    reset_fresh(client, app.config["SECURITY_FRESHNESS"])
     response = client.get("/fresh", headers=headers)
     assert response.status_code == 401
     assert response.json["response"]["reauth_required"]
 
     response = client.get("/verify")
-    assert b"Please Enter Your Password" in response.data
+    assert b"Please Reauthenticate" in response.data
 
     response = client.post(
         "/verify", json=dict(password="not my password"), headers=headers
@@ -1050,6 +1081,36 @@ def test_verify_next(app, client, get_message):
     assert response.location == "http://127.0.0.1:5000/dashboard/settings/"
 
 
+@pytest.mark.webauthn()
+@pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil)
+def test_verify_wan(app, client, get_message):
+    # test get correct options when requiring a reauthentication and have wan keys
+    # setup.
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+
+    reg_2_keys(client)
+
+    reset_fresh(client, app.config["SECURITY_FRESHNESS"])
+    response = client.get("/fresh", headers=headers)
+    assert response.status_code == 401
+    assert response.json["response"]["reauth_required"]
+    assert response.json["response"]["has_webauthn_verify_credential"]
+
+    # the verify form should have the webauthn verify form attached
+    response = client.get("verify")
+    assert b'action="/wan-verify"' in response.data
+
+    app.config["SECURITY_WAN_ALLOW_AS_VERIFY"] = None
+    response = client.get("/fresh", headers=headers)
+    assert response.status_code == 401
+    assert response.json["response"]["reauth_required"]
+    assert not response.json["response"]["has_webauthn_verify_credential"]
+
+    # the verify form should NOT have the webauthn verify form attached
+    response = client.get("verify")
+    assert b'action="/wan-verify"' not in response.data
+
+
 def test_direct_decorator(app, client, get_message):
     """Test/show calling the auth_required decorator directly"""
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
@@ -1157,6 +1218,7 @@ def test_nodatastore(app):
         s.init_app(app)
 
 
+@pytest.mark.filterwarnings("ignore:.*Replacing login_manager.*:DeprecationWarning")
 def test_reuse_security_object(sqlalchemy_datastore):
     # See: https://github.com/Flask-Middleware/flask-security/issues/518
     # Let folks re-use the Security object (mostly for testing).
