@@ -47,13 +47,12 @@ def test_recoverable_flag(app, clients, get_message):
 
     # Test submitting email to reset password creates a token and sends email
     with capture_reset_password_requests() as requests:
-        with app.mail.record_messages() as outbox:
-            response = clients.post(
-                "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
-            )
+        response = clients.post(
+            "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
+        )
 
     assert len(recorded_instructions_sent) == 1
-    assert len(outbox) == 1
+    assert len(app.mail.outbox) == 1
     assert response.status_code == 200
     assert get_message("PASSWORD_RESET_REQUEST", email="joe@lp.com") in response.data
     token = requests[0]["token"]
@@ -149,16 +148,15 @@ def test_recoverable_json(app, client, get_message):
     with capture_flashes() as flashes:
         # Test reset password creates a token and sends email
         with capture_reset_password_requests() as requests:
-            with app.mail.record_messages() as outbox:
-                response = client.post(
-                    "/reset",
-                    json=dict(email="joe@lp.com"),
-                    headers={"Content-Type": "application/json"},
-                )
-                assert response.headers["Content-Type"] == "application/json"
+            response = client.post(
+                "/reset",
+                json=dict(email="joe@lp.com"),
+                headers={"Content-Type": "application/json"},
+            )
+            assert response.headers["Content-Type"] == "application/json"
 
         assert len(recorded_instructions_sent) == 1
-        assert len(outbox) == 1
+        assert len(app.mail.outbox) == 1
         assert response.status_code == 200
         token = requests[0]["token"]
 
@@ -169,9 +167,9 @@ def test_recoverable_json(app, client, get_message):
             headers={"Content-Type": "application/json"},
         )
         assert response.status_code == 400
-        assert response.json["response"]["errors"]["email"][0].encode(
-            "utf-8"
-        ) == get_message("USER_DOES_NOT_EXIST")
+        assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+            "USER_DOES_NOT_EXIST"
+        )
 
         # Test submitting a new password but leave out 'confirm'
         response = client.post(
@@ -180,9 +178,9 @@ def test_recoverable_json(app, client, get_message):
             headers={"Content-Type": "application/json"},
         )
         assert response.status_code == 400
-        assert response.json["response"]["errors"]["password_confirm"][0].encode(
-            "utf-8"
-        ) == get_message("PASSWORD_NOT_PROVIDED")
+        assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+            "PASSWORD_NOT_PROVIDED"
+        )
 
         # Test submitting a new password
         response = client.post(
@@ -227,7 +225,7 @@ def test_recoverable_json(app, client, get_message):
             json=dict(password="newpassword", password_confirm="newpassword"),
             headers={"Content-Type": "application/json"},
         )
-        assert response.json["response"]["error"].encode("utf-8") == get_message(
+        assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
             "INVALID_RESET_PASSWORD_TOKEN"
         )
     assert len(flashes) == 0
@@ -238,10 +236,10 @@ def test_recoverable_template(app, client, get_message):
     # in order to check all context vars since the default template
     # doesn't have all of them.
     with capture_reset_password_requests() as resets:
-        with app.mail.record_messages() as outbox:
-            response = client.post(
-                "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
-            )
+        response = client.post(
+            "/reset", data=dict(email="joe@lp.com"), follow_redirects=True
+        )
+        outbox = app.mail.outbox
         assert len(outbox) == 1
         matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
         # should be 4 - link, email, token, config item
@@ -584,7 +582,7 @@ def test_password_normalization(app, client, get_message):
     assert response.status_code == 200
     logout(client)
 
-    # make sure can log in with new password both normnalized or not
+    # make sure can log in with new password both normalized or not
     response = client.post(
         "/login",
         json=dict(email="matt@lp.com", password="HöheHöhe"),
@@ -603,3 +601,17 @@ def test_password_normalization(app, client, get_message):
     # verify actually logged in
     response = client.get("/profile", follow_redirects=False)
     assert response.status_code == 200
+
+
+@pytest.mark.settings(return_generic_responses=True)
+def test_generic_response(app, client, get_message):
+
+    # try unknown user
+    response = client.post("/reset", data=dict(email="whoami@test.com"))
+    assert (
+        get_message("PASSWORD_RESET_REQUEST", email="whoami@test.com") in response.data
+    )
+
+    response = client.post("/reset", json=dict(email="whoami@test.com"))
+    assert response.status_code == 200
+    assert not any(e in response.json["response"].keys() for e in ["error", "errors"])
