@@ -19,7 +19,7 @@ from flask_security.forms import (
     StringField,
     _default_field_labels,
 )
-from flask_security.signals import user_registered
+from flask_security.signals import user_registered, user_not_registered
 from flask_security.utils import localize_callback
 
 pytestmark = pytest.mark.registerable()
@@ -37,6 +37,7 @@ def test_registerable_flag(clients, app, get_message):
     # Test registering is successful, sends email, and fires signal
     @user_registered.connect_via(app)
     def on_user_registered(app, **kwargs):
+
         assert isinstance(app, Flask)
         assert isinstance(kwargs["user"], UserMixin)
         assert kwargs["confirm_token"] is None
@@ -50,11 +51,10 @@ def test_registerable_flag(clients, app, get_message):
         password_confirm="battery staple",
         next="",
     )
-    with app.mail.record_messages() as outbox:
-        response = clients.post("/register", data=data, follow_redirects=True)
+    response = clients.post("/register", data=data, follow_redirects=True)
 
     assert len(recorded) == 1
-    assert len(outbox) == 1
+    assert len(app.mail.outbox) == 1
     assert b"Post Register" in response.data
 
     logout(clients)
@@ -116,9 +116,9 @@ def test_registerable_flag(clients, app, get_message):
 
 @pytest.mark.confirmable()
 @pytest.mark.babel()
-@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_xlation(app, client, get_message_local):
     # Test form and email translation
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     response = client.get("/register", follow_redirects=True)
@@ -128,16 +128,16 @@ def test_xlation(app, client, get_message_local):
         submit = localize_callback(_default_field_labels["register"])
         assert f'value="{submit}"'.encode() in response.data
 
-    with app.mail.record_messages() as outbox:
-        response = client.post(
-            "/register",
-            data={
-                "email": "me@fr.com",
-                "password": "new strong password",
-                "password_confirm": "new strong password",
-            },
-            follow_redirects=True,
-        )
+    response = client.post(
+        "/register",
+        data={
+            "email": "me@fr.com",
+            "password": "new strong password",
+            "password_confirm": "new strong password",
+        },
+        follow_redirects=True,
+    )
+    outbox = app.mail.outbox
 
     with app.test_request_context():
         assert (
@@ -158,7 +158,7 @@ def test_xlation(app, client, get_message_local):
                     )
                 )
             )
-            in outbox[0].html
+            in outbox[0].alternatives[0][0]
         )
 
 
@@ -197,6 +197,7 @@ def test_required_password_confirm(client, get_message):
 
 @pytest.mark.confirmable()
 @pytest.mark.unified_signin()
+@pytest.mark.settings(password_required=False)
 def test_allow_null_password(client, get_message):
     # If unified sign in is enabled - should be able to register w/o password
     data = dict(email="trp@lp.com", password="")
@@ -205,6 +206,7 @@ def test_allow_null_password(client, get_message):
 
 
 @pytest.mark.unified_signin()
+@pytest.mark.settings(password_required=False)
 def test_allow_null_password_nologin(client, get_message):
     # If unified sign in is enabled - should be able to register w/o password
     # With confirmable false - should be logged in automatically upon register.
@@ -251,9 +253,8 @@ def test_disable_register_emails(client, app):
     data = dict(
         email="dude@lp.com", password="password", password_confirm="password", next=""
     )
-    with app.mail.record_messages() as outbox:
-        client.post("/register", data=data, follow_redirects=True)
-    assert len(outbox) == 0
+    client.post("/register", data=data, follow_redirects=True)
+    assert not app.mail.outbox
 
 
 @pytest.mark.two_factor()
@@ -284,7 +285,7 @@ def test_two_factor_json(app, client, get_message):
     # make sure not logged in
     response = client.get("/profile", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "UNAUTHENTICATED"
     )
 
@@ -301,6 +302,7 @@ def test_form_data_is_passed_to_user_registered_signal(app, sqlalchemy_datastore
 
     @user_registered.connect_via(app)
     def on_user_registered(app, **kwargs):
+
         assert isinstance(app, Flask)
         assert isinstance(kwargs["user"], UserMixin)
         assert kwargs["confirm_token"] is None
@@ -347,7 +349,7 @@ def test_easy_password(app, sqlalchemy_datastore):
     assert response.headers["Content-Type"] == "application/json"
     assert response.status_code == 400
     # Response from zxcvbn
-    assert "Repeats like" in response.json["response"]["errors"]["password"][0]
+    assert "Repeats like" in response.json["response"]["errors"][0]
 
     # Test that username affects password selection
     data = dict(
@@ -364,7 +366,7 @@ def test_easy_password(app, sqlalchemy_datastore):
     # Response from zxcvbn
     assert (
         "Password not complex enough"
-        in response.json["response"]["errors"]["password"][0]
+        in response.json["response"]["field_errors"]["password"][0]
     )
 
 
@@ -390,9 +392,9 @@ def test_nullable_username(app, client):
 
 
 def test_email_normalization(app, client):
-    # should be able to login either as LP.com or lp.com
+    # should be able to login either as LP.com or lp.com or canonical unicode form
     data = dict(
-        email="\N{BLACK SCISSORS}@LP.com",
+        email="Imnumber\N{OHM SIGN}@LP.com",
         password="battery staple",
         password_confirm="battery staple",
     )
@@ -403,14 +405,23 @@ def test_email_normalization(app, client):
 
     # Test user can login after registering
     response = authenticate(
-        client, email="\N{BLACK SCISSORS}@lp.com", password="battery staple"
+        client, email="Imnumber\N{OHM SIGN}@lp.com", password="battery staple"
     )
     assert response.status_code == 302
 
     logout(client)
     # Test user can login after registering using original non-canonical email
     response = authenticate(
-        client, email="\N{BLACK SCISSORS}@LP.com", password="battery staple"
+        client, email="Imnumber\N{OHM SIGN}@LP.com", password="battery staple"
+    )
+    assert response.status_code == 302
+
+    logout(client)
+    # Test user can login after registering using original non-canonical email
+    response = authenticate(
+        client,
+        email="Imnumber\N{GREEK CAPITAL LETTER OMEGA}@LP.com",
+        password="battery staple",
     )
     assert response.status_code == 302
 
@@ -448,8 +459,6 @@ def test_form_error(app, client, get_message):
     from babel.support import LazyProxy
 
     with app.test_request_context("/register"):
-        # this is usually done @before_first_request
-        app.jinja_env.globals["_fsdomain"] = app.security.i18n_domain.gettext
         rform = ConfirmRegisterForm()
 
         rform.validate()
@@ -481,12 +490,37 @@ def test_username(app, client, get_message):
     assert response.status_code == 200
     logout(client)
 
-    # login using historic - email field...
+    # login using historic - email field to hold a username - won't work since
+    # it is an EmailField
     response = json_authenticate(client, email="dude", password="awesome sunset")
+    assert response.status_code == 400
+    assert (
+        get_message("INVALID_EMAIL_ADDRESS", username="dude")
+        == response.json["response"]["errors"][0].encode()
+    )
+    # login using username
+    response = client.post(
+        "/login", json=dict(username="dude", password="awesome sunset")
+    )
     assert response.status_code == 200
-
     logout(client)
 
+    # login with email
+    response = client.post(
+        "/login", json=dict(email="dude@lp.com", password="awesome sunset")
+    )
+    assert response.status_code == 200
+    logout(client)
+
+    response = client.post(
+        "/login", json=dict(emails="dude", password="awesome sunset")
+    )
+    assert response.status_code == 400
+    assert (
+        get_message("USER_DOES_NOT_EXIST")
+        == response.json["response"]["field_errors"]["null"][0].encode()
+    )
+
     # login using us-signin
     response = client.post(
         "/us-signin",
@@ -508,7 +542,7 @@ def test_username(app, client, get_message):
     assert response.status_code == 400
     assert (
         get_message("USERNAME_ALREADY_ASSOCIATED", username="dude")
-        == response.json["response"]["errors"]["username"][0].encode()
+        == response.json["response"]["field_errors"]["username"][0].encode()
     )
 
 
@@ -547,20 +581,14 @@ def test_username_errors(app, client, get_message):
         "/register", json=data, headers={"Content-Type": "application/json"}
     )
     assert response.status_code == 400
-    assert (
-        "Username must be at least"
-        in response.json["response"]["errors"]["username"][0]
-    )
+    assert "Username must be at least" in response.json["response"]["errors"][0]
 
     data["username"] = "howlongcanIbebeforeyoucomplainIsupposereallyreallylong"
     response = client.post(
         "/register", json=data, headers={"Content-Type": "application/json"}
     )
     assert response.status_code == 400
-    assert (
-        "Username must be at least"
-        in response.json["response"]["errors"]["username"][0]
-    )
+    assert "Username must be at least" in response.json["response"]["errors"][0]
 
     # try evil characters
     data["username"] = "hi <script>doing evil</script>"
@@ -570,7 +598,7 @@ def test_username_errors(app, client, get_message):
     assert response.status_code == 400
     assert (
         get_message("USERNAME_ILLEGAL_CHARACTERS")
-        == response.json["response"]["errors"]["username"][0].encode()
+        == response.json["response"]["field_errors"]["username"][0].encode()
     )
 
     # No spaces or punctuation
@@ -581,7 +609,7 @@ def test_username_errors(app, client, get_message):
     assert response.status_code == 400
     assert (
         get_message("USERNAME_DISALLOWED_CHARACTERS")
-        in response.json["response"]["errors"]["username"][0].encode()
+        in response.json["response"]["errors"][0].encode()
     )
 
     data["username"] = None
@@ -591,7 +619,7 @@ def test_username_errors(app, client, get_message):
     assert response.status_code == 400
     assert (
         get_message("USERNAME_NOT_PROVIDED")
-        == response.json["response"]["errors"]["username"][0].encode()
+        == response.json["response"]["errors"][0].encode()
     )
 
 
@@ -599,3 +627,220 @@ def test_username_not_enabled(app, client, get_message):
     response = client.get("/register")
     assert b"username" not in response.data
     assert not hasattr(RegisterForm, "username")
+
+
+def test_legacy_style_login(app, sqlalchemy_datastore, get_message):
+    # Show how to setup LoginForm to mimic legacy behavior of
+    # allowing any identity in the email field.
+    # N.B. for simplicity we don't enable confirmable....
+    from flask_security import (
+        RegisterForm,
+        LoginForm,
+        Security,
+        uia_username_mapper,
+        unique_identity_attribute,
+    )
+    from flask_security.utils import lookup_identity
+    from werkzeug.local import LocalProxy
+    from wtforms import StringField, ValidationError, validators
+
+    def username_validator(form, field):
+        # Side-effect - field.data is updated to normalized value.
+        # Use proxy to we can declare this prior to initializing Security.
+        _security = LocalProxy(lambda: app.extensions["security"])
+        msg, field.data = _security._username_util.validate(field.data)
+        if msg:
+            raise ValidationError(msg)
+
+    class MyRegisterForm(RegisterForm):
+        # Note that unique_identity_attribute uses the defined field 'mapper' to
+        # normalize. We validate before that to give better error messages and
+        # to set the normalized value into the form for saving.
+        username = StringField(
+            "Username",
+            validators=[
+                validators.data_required(),
+                username_validator,
+                unique_identity_attribute,
+            ],
+        )
+
+    class MyLoginForm(LoginForm):
+        email = StringField("email", [validators.data_required()])
+
+        def validate(self, **kwargs):
+            self.user = lookup_identity(self.email.data)
+            self.ifield = self.email
+            if not super().validate(**kwargs):
+                return False
+            return True
+
+    app.config["SECURITY_USER_IDENTITY_ATTRIBUTES"] = [
+        {"username": {"mapper": uia_username_mapper}}
+    ]
+    security = Security(
+        datastore=sqlalchemy_datastore,
+        register_form=MyRegisterForm,
+        login_form=MyLoginForm,
+    )
+    security.init_app(app)
+
+    client = app.test_client()
+
+    data = dict(
+        email="mary2@lp.com",
+        username="mary",
+        password="awesome sunset",
+        password_confirm="awesome sunset",
+    )
+    response = client.post("/register", data=data, follow_redirects=True)
+    assert b"Welcome mary" in response.data
+    logout(client)
+
+    # log in with username
+    response = client.post(
+        "/login",
+        data=dict(email="mary", password="awesome sunset"),
+        follow_redirects=True,
+    )
+    # verify actually logged in
+    response = client.get("/profile", follow_redirects=False)
+    assert response.status_code == 200
+
+
+@pytest.mark.confirmable()
+@pytest.mark.settings(return_generic_responses=True, username_enable=True)
+def test_generic_response(app, client, get_message):
+    recorded = []
+
+    @user_not_registered.connect_via(app)
+    def on_user_registered(app, **kwargs):
+        recorded.append(kwargs)
+
+    # Register should not expose whether email/username is already in system.
+    # Should still return errors such as illegal password, ...
+    data = dict(
+        email="dude@lp.com",
+        username="dude",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 200
+    assert len(app.mail.outbox) == 1
+    assert len(recorded) == 0
+
+    # try again - should not get ANY error - but should get an email
+    response = client.post("/register", json=data)
+    assert response.status_code == 200
+    assert not any(
+        e in response.json["response"].keys() for e in ["errors", "field_errors"]
+    )
+    gr = app.mail.outbox[1]
+    assert "tried to register this email" in gr.body
+    assert "associated with it: dude" in gr.body
+    assert len(recorded) == 1
+    # test that signal sent.
+    nr = recorded[0]
+    assert nr["existing_email"]
+    assert nr["user"]
+    assert nr["form_data"]["email"] == "dude@lp.com"
+
+    # Forms should get generic response - even though email already registered.
+    response = client.post("/register", data=data, follow_redirects=True)
+    assert get_message("CONFIRM_REGISTRATION", email="dude@lp.com") in response.data
+    assert len(app.mail.outbox) == 3
+    assert len(recorded) == 2
+
+    # Try same email with different username
+    response = client.post(
+        "/register",
+        data=dict(email="dude@lp.com", username="dude2", password="awesome sunset"),
+        follow_redirects=True,
+    )
+    assert get_message("CONFIRM_REGISTRATION", email="dude@lp.com") in response.data
+    assert len(app.mail.outbox) == 4
+    assert len(recorded) == 3
+    gr = app.mail.outbox[3]
+    assert "tried to register this email" in gr.body
+    assert "associated with it: dude" in gr.body
+
+    # Now test a new email with an existing username
+    response = client.post(
+        "/register",
+        data=dict(email="dude39@lp.com", username="dude", password="awesome sunset"),
+        follow_redirects=True,
+    )
+    assert get_message("CONFIRM_REGISTRATION", email="dude39@lp.com") in response.data
+    assert len(app.mail.outbox) == 5
+    assert len(recorded) == 4
+    gr = app.mail.outbox[4]
+    assert 'You attempted to register with a username "dude" that' in gr.body
+    # test that signal sent.
+    nr = recorded[3]
+    assert not nr["existing_email"]
+    assert not nr["user"]
+    assert nr["existing_username"]
+    assert nr["form_data"]["username"] == "dude"
+
+    # should still get detailed errors about e.g. bad password
+    data = dict(
+        email="dude@lp.com",
+        username="dude",
+        password="a",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 400
+    assert response.json["response"]["errors"][0].encode() == get_message(
+        "PASSWORD_INVALID_LENGTH", length=8
+    )
+    data = dict(
+        email="dude@lp.com",
+        username="dd",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 400
+    assert response.json["response"]["errors"][0].encode() == get_message(
+        "USERNAME_INVALID_LENGTH", min=4, max=32
+    )
+
+
+@pytest.mark.recoverable()
+@pytest.mark.confirmable()
+@pytest.mark.settings(return_generic_responses=True, username_enable=True)
+def test_generic_response_recover(app, client, get_message):
+    # If user tries to re-register - response email should contain recovery url
+    recorded = []
+
+    @user_not_registered.connect_via(app)
+    def on_user_registered(app, **kwargs):
+        recorded.append(kwargs)
+
+    data = dict(
+        email="dude@lp.com",
+        username="dude",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 200
+    assert len(app.mail.outbox) == 1
+    assert len(recorded) == 0
+
+    # now same email - same or different username
+    data = dict(
+        email="dude@lp.com",
+        username="dude2",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 200
+    assert len(app.mail.outbox) == 2
+    gr = app.mail.outbox[1]
+    assert "/reset" in gr.body
+
+    assert len(recorded) == 1
+    # test that signal sent.
+    nr = recorded[0]
+    assert nr["existing_email"]
+    assert nr["user"]
+    assert nr["form_data"]["email"] == "dude@lp.com"
